#!/usr/bin/python3
#
# Hilly cipher by Yahya. This cipher is an extension of the Hill cipher.
#
# This program was developed by Dennis Brinkrolf using Python 3.5.2
# with Eclipse (pydev) as IDE under Windows and Linux.
# 
# Some command line calls with some typical options and arguments:
#
# - python Hilly.py encrypt -q -i plaintext.txt -k key.txt
#     Uses Hilly to encrypt plaintext with a key file and writes the ciphertext
#     to the console (because of -q the console output is reduced)
#
# - python Hilly.py generate -n 7 -x 200 -y 300 -f 1000000 -o key.txt  
#     Generates a 7x7 key matrix in 1000000 tries for Hilly with min integer
#     values -200 and max 300, and saves the key matrix to file.
#
# - python Hilly.py generate -o C:\Users\XY\Desktop\key.txt 
#     Generates a 6x6 key matrix and saves the key matrix to Desktop (on Windows).
#   python Hilly.py generate -o /home/XY/key.txt
#     Generates a 6x6 key matrix and saves the key matrix to a file (on Linux:).
#
# - python Hilly.py encrypt -e -i plaintext.txt -k key.txt -o ciphertext.txt 
#     Uses usual Hill to encrypt the plaintext with a key file
#     and saves the ciphertext to the specified file.
#
# - python Hilly.py encrypt -i plaintext.txt -k key.txt -o ciphertext.txt 
#     Uses Hilly to encrypt the plaintext with a key file
#     and saves the ciphertext to the specified file.
#
# - python Hilly.py decrypt -i ciphertext.txt -k key.txt -o plaintext.txt
#     Uses Hilly to decrypt the ciphertext with a key file
#     and saves the plaintext to the specified file.
#
# Remark:
# The Python "NumPy" package must be installed to use this script.
# It can easily be installed on the command line with: pip install numpy
#

import sys
import argparse
import string
import numpy as np
from io import StringIO

version = "version 1.04 (2017-01-20)"

#TextIOWrapper
logg_it = StringIO()

#Flag if -v is used
verbose_output = False

#Matrix size n x n
default_matrix_size = 6
matrix_size = default_matrix_size

#Min integer values of generated key matrix
min_val_key = -190

#Max integer values of generated key matrix
max_val_key = 190

#Flag for key file
key_file = False

#Store the key and inverse key matrix information
key_matrix = None
key_matrix_inverse = None

#Amount of tries to generate key file
deeps = 1000

#Amount of padding
pad_size = 0

#Input string
in_s = None

#Output path
path_o = None

#Store p vector for every block
p_blocks = list()

#Store plaintext for every block
plaintext_blocks = list()

#Store ciphertext for every block
cipher_blocks = list()

#Ciphertext substitution table
sub_table_c = list()

#Plaintext substitution table used in the Hill cipher
hill_sub_table_p = list()


#Global variables for Hilly cipher

#Flag to check if Hilly is activated (default 0 means to use Hilly cipher; 1 means to use Hill cipher)
d_opt = 0 

#Sum of the key matrix diagonal
hilly_s = 0

#Flag if hilly_s should be set to 1 in Hilly cipher (option -s) (like in Hill)
flag_hilly_s = 0

#Flag, which allows to test insecure Hilly key matrices (condition 'invertible det(K) mod 26 fulfilled)
hilly_vuln = False

#Plaintext substitution table T_P used in the Hilly cipher (differs by diagonal sum s as multiplier from hill_sub_table_p)
hilly_sub_table_p = {}

#Additional substitution table T_A used in the Hilly cipher
hilly_sub_table_add = list()

#User specific substitution table T_A
hilly_user_sub = None

hilly_sub_table_add_minus = ''

#Hilly extension matrix E
hilly_e_matrix = None

#Hilly inverse extension matrix E^-1
hilly_e_matrix_inverse = None

#Store C1 (C not modulo reduced) for every block
hilly_c1_blocks = list()

#Store extension X for every block
hilly_extensions = list()

#Store encrypted extension X_ENC for every block
hilly_x_enc_blocks = list()

#Store Hilly ciphertext version for every block
hilly_cipher_blocks_extended = list()


'''
This function redirects the console output (sys.stdout) to
TextIOWrapper. 
'''
def activate_logger():
    global logg_it
    sys.stdout = logg_it

'''
This function reset the console output to default sys.stdout.
'''
def deactivate_logger():
    sys.stdout = sys.__stdout__
'''
This function reset the console output and close the global TextIOWrapper logg_it.
'''
def close_logger():
    global logg_it
    
    if sys.stdout != sys.__stdout__:
        deactivate_logger()
    logg_it.close()

'''
This function is only needed for output at console.
'''
def line():
    print("-----------------------------------------------------")

'''
This function implements the Extended Euclidean algorithm.
Check: https://en.wikibooks.org/wiki/Algorithm_Implementation/Mathematics/Extended_Euclidean_algorithm
'''
def egcd(a, b):
    if a == 0:
        return (b, 0, 1)
    else:
        g, x, y = egcd(b % a, a)
        return (g, y - (b // a) * x, x)

'''
This function computes the multiplicative x of a value b modulo n like (x*b) % n = 1.
return: None, if multiplicative inverse doesn't exist.
'''
def mulinv(b, n):
    g, x, _ = egcd(b, n)
    if g == 1:
        return x % n

'''
This function stores the global variable key_matrix as a string in a specific file.
'''
def save_key_matrix():
    global key_matrix, path_o
    
    #Convert numpy ndarray to list
    key_matrix = list(key_matrix)
    
    #Initialize output string
    output = ''
    
    #Iterate key matrix
    for i in range(len(key_matrix)):
        #Get row i of key matrix
        row = key_matrix[i]
        #Concatenate every row
        output = output + '\t'.join(str(x) for x in row) + '\n'
        
    #Write key matrix as string in a specific file
    save_file("\n"+output)

'''
This function writes a string and uses the global variable path_o to store the string in a specific file.
Parameter: val    the string to write
'''                          
def save_file(val): 
    global path_o 
    
    #Check if global variable path_o is set
    if path_o == None:
        sys.exit("No output file used.")
    
    #Open file in write mode
    f = open(path_o,'w')

    #Write the value to file
    f.write(val)
    
    #Output
    print("Write: \"",val,"\" \nto file: ",path_o,sep='')
    
    #Close file
    f.close()

'''
This function reads a file and returns it as string.
Parameter: file      path to file
Return:    str_read  read string
'''
def read_f(file): 
    if file == None:
        sys.exit("No file to open.")
        
    #Open file in read mode
    f = open(file,'r')
    
    #Read file
    str_read = f.read()
    
    #Close file
    f.close()
    
    #Remove all '\n' characters from the end
    str_read = str_read.strip('\n')
    return str_read

'''
This function checks the conditions for the Hill and the Hilly key matrix.
Return: True   key matrix is valid
        False  key matrix is not valid
'''
def key_matrix_check():
    global key_matrix,s,hilly_vuln
    
    
    #Compute long determinant
    sign, logdet = np.linalg.slogdet(key_matrix)    
    det = sign * np.exp(logdet)
    
    #Round and convert as int64
    det = np.round(det,decimals=0).astype(np.int64)
    
    #Need positive determinant for invert process
    if det < 0:
        det = int(det)
        det = abs(det)
    if det == 0:
        #Matrix is not invertible
        return False 
    
    #Compute inverse of determinant
    inv_det = mulinv(det, 26)
    
    #Check which cipher is used
    if d_opt != 1: #Hilly is activated
        global hilly_s
        #Determinant is NOT invertible modulo 26 is a condition which has to be met for Hilly cipher.
        #If vuln flag is set, we accept this nevertheless for testing purposes only.

        if inv_det == None:
            if verbose_output == True:
                print("For Hilly: Good that determinant ",det," is not invertible mod 26.",sep='')
            
            #Compute s for later use
            hilly_s = hilly_calc_s()
            
            if hilly_s == 0:
                if verbose_output == True:
                    print("New matrix needs to be calculated as hilly_s==0.")
                #We need a new key matrix
                return False
            else:
                return True
        else:
            if hilly_vuln == True:
                
                print("For Hilly: Bad, that condition that det(k) mod 26 is invertible is fulfilled. Allowed just to test insecure Hilly matrices.")
                line()
            
                #Compute s for later use
                hilly_s = hilly_calc_s()
            
                if hilly_s == 0:
                    if verbose_output == True:
                        print("New matrix needs to be calculated as hilly_s==0.")
                    
                    #We need a new key matrix
                    return False
                else:
                    return True
            else:
                return False

    else: #Hill is activated
        #Unlike Hilly, determinant is not invertible modulo 26 is a wrong condition for Hill Cipher.
        #Or positively said: For the Hill cipher the determinant has to be invertible modulo 26.
        if inv_det == None:
            return False
        else:
            if verbose_output == False:
                print("For Hill: Good that determinant ",det," is invertible mod 26 with: ",inv_det,".",sep='')
            return True  

'''
This function tries to generate a key matrix, until a valid one
is found or the maximum amount of iterations is reached. 
Return: True      a valid key matrix is found.
'''
def gen_key_matrix_tries():
    global deeps
    
    counter = 0
    while(counter < deeps):
        flag = gen_key_matrix()
        if flag == True:
            return True
        counter = counter + 1
        if counter == deeps:
            sys.exit("More tries than: " + str(deeps) + " are needed to generate key matrix, use option -f.")
    

'''
This function generates the key matrix for Hilly and Hill ciphers and
stores the results in the global variable key_matrix.
Return: true    key matrix is valid
        false   key matrix is not valid
'''
def gen_key_matrix():
    global key_matrix,min_val_key,max_val_key,matrix_size,key_file
    
    #No key file used, generate new key matrix
    if key_file == False:
        #Generate random nxn matrix
        if min_val_key < max_val_key:
            key_matrix = np.random.randint(min_val_key,max_val_key, size=(matrix_size, matrix_size))
        else:
            sys.exit("Generating a key matrix with integer values min > max is not possible.")
        #Check if key matrix is valid
        if key_matrix_check() == True:
            print("\nGenerate ",matrix_size,"x",matrix_size, " key matrix with integer values between (", min_val_key,",",max_val_key,"):\n",key_matrix,sep='')
            return True
        else:
            return False
    #Read key file here
    else:
        #Split the key file in x rows
        tmp = key_file.split('\n')
        #Define matrix_size (n)
        matrix_size = len(tmp)
        
        #Remove tabulator from every row and append to list rows
        rows = list()
        for j in range(len(tmp)):
            tmp[j] = tmp[j].strip('\t')
            rows.append((tmp[j].split('\t')))

        #Initialize 1xn key matrix and fill with zeroes
        dim = (1,matrix_size)
        key_matrix = np.zeros(dim)
        
        #Iterate key matrix
        for i in range(matrix_size):
            #Insert read rows from key file in the key matrix
            key_matrix = np.insert(key_matrix,i,rows[i],axis=0)

        #Delete the last row because it only contains 0 and is not needed
        key_matrix = np.delete(key_matrix,-1,axis=0)
        
        #Convert key matrix to integer
        key_matrix = key_matrix.astype(np.int64)
        
        #Check if key matrix is valid
        if key_matrix_check() == False:
            if d_opt != 1:
                sys.exit("Key matrix is not valid for Hilly cipher.")
            else:
                sys.exit("Key matrix is not valid for Hill cipher.")
        if verbose_output == False:
            deactivate_logger()
            print("Read key matrix from file ",matrix_size,"x",matrix_size,":\n",key_matrix,sep='')
            line()
            activate_logger()
        else:
            print("\nRead key matrix from file ",matrix_size,"x",matrix_size,":\n",key_matrix,sep='')
        return True

'''
This function checks if padding is needed. If so if adds 'Z's to the list x such that length(x) divides block length (bl).
Parameter: x    list like plaintext or ciphertext
           bl   block length
Return: x    padded list
'''
def check_plain_padding(x,bl):
    global pad_size,matrix_size
    
    #Get length
    l = len(x)
    
    #Convert to list
    x = list(x)
    
    #Compute padding size
    pad_size = bl - (l % bl)
    
    #Check if padding needed
    if pad_size == 0 or pad_size == matrix_size:
        print("=> no padding needed.")
    else:
        #Fill with 'Z' till pad_size is reached
        print("=> we need padding with ",pad_size," values and letter: Z",sep='')
        #Padding with list till i is reached with a specific character
        for l in range(pad_size):
            x.append('Z')

    #Return padded list
    return x
    
'''
This function removes padding from a list
Parameter: x    list with padded values
           c    amount of values to remove
Return: x     list without padding
'''
def remove_padding(x,c):
    print("Deconstruct padding, remove the last",c,"chars.")
    
    #Return list x without the last c values
    return x[:-c]

'''
This function splits a given list x into blocks of the given block length bl.
Parameter: x    list to split
           bl   block length
Return: split_done    list of blocks
'''
def split_it(x,bl):
    
    #Get length of the given list
    block_l = len(x)
        
    #Compute the amount of blocks
    anz_blocks = int(block_l /bl)
    
    print("Amount of blocks:",anz_blocks)
    
    #Initialize variables
    split_done = list()
    start = 0 
    end = bl
   
    #Iterate blocks
    for i in range(anz_blocks):
        
        #Compute window (subset of the list)
        window = x[start:end]
        
        #Append to temporary list
        split_done.append(window)
        
        print("Block ",i,":\n",window,sep='')
        
        #Compute next window
        start = end
        end = end + bl
    line()
    #Return the blocks as list
    return split_done

'''
This function computes substituted C_i and C1_i blocks (if Hilly is used) and 
stores the results in the global list hilly_c1_blocks and cipher_blocks.
Parameter: p_j    p vector block
           j      only used for output
'''  
def create_c_i(p_j,j):
    global d_opt,cipher_blocks,sub_table_c
    
    #Compute C1_j = K * P_j  
    c1 = np.dot(key_matrix,p_j).astype(np.int64)
    
    #C1_j blocks only needed for Hilly 
    if d_opt != 1:
        global hilly_c1_blocks
        
        #Append C1_j (C_j not reduced modulo) in global list for Hilly
        hilly_c1_blocks.append(c1)
        print("Compute C1_",j," = K * P_",j,":\n",list(c1),"\n",sep='')
    
    #Compute C1_j mod 26
    c_tmp = np.mod(c1,26)
    print("Compute C_",j," = K * P_",j," mod 26:\n",list(c_tmp),sep='')
    
    #Initialize list
    tmp_l = list()
    
    #Iterate C_j and use substitution table (T_C)
    for i in range(len(c_tmp)):
        #Append the substituted value in the temporary list
        tmp_l.append(sub_table_c[c_tmp[i]])

    print("\nUse substitution table (T_C) for C_",j,":\n",tmp_l,sep='')
    
    #Append substituted C_j vector to global list cipher_blocks
    cipher_blocks.append(tmp_l)

'''
This function generates plaintext and ciphertext substitution tables for the Hill cipher.
The substitution table for plaintext (T_P) is stored in the global list hill_sub_table_p.
The substitution table for ciphertext (T_C) is stored in the global list sub_table_c.
Remark: Compare to function hilly_gen_sub_tables() for the Hilly cipher.
'''
def hill_gen_sub_tables():
    global hill_sub_table_p,sub_table_c
    
    print("Generating substitution tables..")
    
    #Fill hill_sub_table_p list with A-Z
    print("\nPlaintext substitution table (T_P):")
    for i in range(26):
        
        #Get Ascii uppercase at position i
        letter = string.ascii_uppercase[i]
        
        #Append to global list hill_sub_table_p
        hill_sub_table_p.append(letter)
        print("(",i,",",letter,")",sep='',end=' ')
    print("")
    
    #Fill sub_table_c list with A-Z
    print("\nCiphertext substitution table (T_C):")  
    for j in range(26):
        #Get ASCII uppercase at position j
        letter = string.ascii_uppercase[j]
        #Append to global list sub_table_c
        sub_table_c.append(letter)
        print("(",j,",",letter,")",sep='',end=' ')
    print("")

'''
This function uses the substitution table for plaintext (T_P) and builds the plaintext vector P_j. 
Furthermore it stores the result in the global list p_blocks.
Parameter: plain_j    plaintext block j
           j          only used for output
'''
def hill_enc_sub_p_i(plain_j,j):
    global p_blocks,hill_sub_table_p
        
    #Initialize list which stores the p vector block j
    p_tmp = list()
    
    #Iterate plaintext block
    for i in range(len(plain_j)):
        #Get plaintext char i
        val_c = plain_j[i]
        #Use substitution table for plaintext and get int value of plaintext char
        val_int = hill_sub_table_p.index(val_c)
        
        #Append the substituted plaintext char to list
        p_tmp.append(val_int)
            
    #Append the p vector block j to the global list
    p_blocks.append(p_tmp)
        
    print("Plaintext vector P_",j,":\n",p_tmp,sep='')

'''
This function inverts the key matrix for Hilly and stores the result 
in the global variable key_matrix_inverse.
'''
def hill_dec_gen_invert():
    global key_matrix,key_matrix_inverse
    
    print("Computing inverted key matrix..\n")
   
    #Compute determinant of key matrix
    sign, logdet = np.linalg.slogdet(key_matrix)
    det = sign * np.exp(logdet)
    
    #Round determinant and convert as int
    det = np.round(det, decimals=0).astype(np.int64)
    print("Determinant of key matrix:\n",det,sep='')
        
    #Compute inverse key matrix (expanded form)
    key_inverse_expanded = np.linalg.inv(key_matrix)
    
    #Compute positive determinant
    if det < 0:
        det = abs(det)

    #Compute inverse determinant 
    invert_det = mulinv(det, 26)
    print("\nComputing inverse determinant det(K)^-1 mod 26:\n",invert_det,sep='')
    
    #Initialize variable
    key_inverse_normal = key_inverse_expanded
    
    #Check if we need to compute normal form
    if det != 1:
        
        #Remove determinant from key inverse matrix (normal form)
        key_inverse_normal = np.dot(key_inverse_expanded,det)
        
        #Round key_inverse_normal and convert as int 
        key_inverse_normal = np.round(key_inverse_normal,decimals=0).astype(np.int64)
    
    #Compute inverse key matrix as K^-1 = det(K)^-1 * K^-1 
    key_inverse_final = np.dot(key_inverse_normal,invert_det)
    
    #Round inverse key matrix and convert as int
    key_inverse_final = np.round(key_inverse_final,decimals=0).astype(np.int64)
    
    print("\nInverted K^-1 = (det(K)^-1 * K^-1):\n",key_inverse_normal,sep='')
    
    #Reducing modulo 26
    key_inverse_final = np.mod(key_inverse_final,26)
    
    print("\nReducing K^-1 mod 26:\n",key_inverse_final,sep='')
    #Store in global variable
    key_matrix_inverse = key_inverse_final

'''
This function uses the substitution table (T_c) to substituting ciphertext block C_j  
and stores the result in the global list cipher_blocks.
Parameter: c_j   ciphertext block j
           j     only used for output
'''
def hill_dec_sub_i(c_j,j):
    global cipher_blocks
    
    #Initialize list
    tmp = list()
    
    #Iterate cipher block c_j
    for i in range(len(c_j)):
        #Get the int value at position i in the cipher block from T_c
        val = sub_table_c.index(c_j[i])
        #Append replaced value to tmp list
        tmp.append(val)
    
    #Append the substituted cipher block in the global list
    cipher_blocks.append(tmp)
    print("After use cipher substitution table (T_c) for C_",j,":\n",tmp,sep='')

'''
This function decrypts the plaintext vector P_i and uses the substitution table for plaintext (T_P).
The results are stored in the global lists p_blocks and plaintext_blocks.
Parameter: c_j   ciphertext block j
           j     only used for output
''' 
def hill_dec_p_i(c_j,j):
    global key_matrix_inverse,p_blocks,hill_sub_table_p,plaintext_blocks
    
    #Compute p_j = K^-1 * C_j mod 26
    p_dec = np.dot(key_matrix_inverse,c_j)
    p_dec = np.mod(p_dec,26)
    print("Decrypting block P_",j," = K^-1 * C_",j,":\n",p_dec,sep='')
    
    #Store decrypted p block in global list
    p_blocks.append(p_dec)
    
    #Initialize list
    p_sub = list()
    
    #Iterate decrypted p block
    for c in range(len(p_dec)):
        
        #Int value at position
        position = p_dec[c]
        
        #Use Hill substitution table for plaintext (T_P) to get char
        plain_char = hill_sub_table_p[position]
        
        #Append to list
        p_sub.append(plain_char)
        
    print("After using T_P:\n",p_sub,sep='')
    
    #Store in global list
    plaintext_blocks.append(p_sub)

'''
This function computes the diagonal sum of the key matrix.
Remark: With the option -s this could be fixed to 1 like in normal Hill cipher.
Return: s    sum of the diagonal
'''
def hilly_calc_s():
    global key_matrix
    
    #Initialize variables
    tmp1 = 0
    s = 0
    
    #Compute sum of the main diagonal
    for i in range(0,matrix_size):
        s = s + key_matrix[i][tmp1]
        tmp1 += 1

    #Check if main diagonal is empty
    if s == 0: 
        #Compute sum of the next diagonal
        tmp2 = matrix_size-1
        for j in range(0,matrix_size):
            s = s + key_matrix[j][tmp2]
            tmp2 -= 1

    #Return sum of diagonal
    return s

'''
This function generates the substitution table for plaintext (T_P) and for ciphertext (T_C)
used in the Hilly cipher.
The results are stored in the global lists hilly_sub_table_p and sub_table_c.
'''           
def hilly_gen_sub_tables():
    global hilly_s,hilly_sub_table_p,sub_table_c
    
    print("Substitution table for plaintext (T_P) with s=",hilly_s,":",sep='')
    
    #Initialize variable
    count = 1
    
    #Iterate alphabet 
    for i in range(0,26):
        
        #Get value like 'A'
        c = string.ascii_uppercase[i]

        #Compute the substitution table for plaintext (T_P)
        hilly_sub_table_p[c] = count * hilly_s
        
        #Compute the substitution table for ciphertext (T_C)
        sub_table_c.append(c)
        
        print("(",c,",",hilly_sub_table_p.get(c),")",sep='',end='')
        
        #Increment counter 
        count += 1

    print("")
    print("\nSubstitution table for ciphertext (T_C):")
    
    #Left shift by 1 of sub_table_c and get Z=0, A=1 ... Y=25
    sub_table_c = np.roll(sub_table_c,1)
  
    #Output
    for j in range(26):
        print("(",sub_table_c[j],",",j,")",sep='',end='')
    print("")

'''
This function computes the additional substitution table and stores the result
in the global list hilly_sub_table_add.
'''   
def hilly_gen_sub_table_extended():
    global key_matrix,hilly_sub_table_add,hilly_sub_table_add_minus,hilly_user_sub
    
    tmp = None
    
    #Check if user specific additional subtable is used.
    if hilly_user_sub != None:
        
        #Remove tabs and split the string
        tmp = hilly_user_sub.split("\t")
        
        #Check length of the additional subtable
        if len(tmp) < 11 or len(tmp) > 11:
            print("Additional subtable need 11 special characters.")
            sys.exit()
    
        #Check if alpha charaters are used
        if hilly_user_sub.isalpha() == True:
            print("Only non alpha characters are allowed.")
            sys.exit()

        #Check for duplicate entries
        for x in tmp:
            c = tmp.count(x)
            if c > 1:
                print("No duplicate entries are allowed.")
                sys.exit()
        line()
        if verbose_output == False:
            deactivate_logger()
            print("Read the additional substitution table from a given file:\n", tmp,sep="")
            line()
            activate_logger()
        else:
            print("Read the additional substitution table from a given file..")

    else:
        #Initialize array
        tmp = ['*','#','&','%','$','+','?','!','@','^','-']
    #Compute sum of the key matrix modulo
    t = (np.sum(key_matrix)) % 11
    
    #Check if we need to rotate right
    if t != 0:
        
        #Cast as int
        t = t.astype(int)
        print("\n\nWe need to right shift the default additional substitution table by:",t,"positions.")
          
        #Right shift by t positions and convert to list 
        hilly_sub_table_add = list(np.roll(tmp,t))
    else:
        #No shift needed
        hilly_sub_table_add = list(tmp)
    hilly_sub_table_add_minus = hilly_sub_table_add[0]
    print("\nThe substitution table for additional symbols:\n",hilly_sub_table_add,sep='')
    hilly_sub_table_add = hilly_sub_table_add[1:]
         

'''
This function computes the extension matrix E and stores the result in the global variable hilly_e_matrix.
'''
def hilly_e_init():
    global matrix_size,s,hilly_e_matrix,hilly_s
    
    #Initialize numpy ndarray of type int 
    s_hilly_e_matrix = np.ndarray(shape=(matrix_size,matrix_size), dtype=int)
    
    #Fill the ndarry with the global variable s
    s_hilly_e_matrix.fill(hilly_s)
    
    #Compute extension matrix, E = s * I + K
    hilly_e_matrix = np.add(s_hilly_e_matrix,key_matrix)
    print("Extension matrix:\n",hilly_e_matrix,sep='')

'''
This function computes for the Hilly encryption mode the substituted plaintext vector P_j and 
stores the result in the global list p_blocks.
Parameter: plain_j   plaintext block j
           j         only used for output
'''
def hilly_enc_sub_p_i(plain_j,j):
    global p_blocks,sub_table_p
    #Initialize list
    tmp_l = list()
    
    #Iterate plaintext block j
    for i in range(len(plain_j)):
        #Use substitution table for plaintext (T_P) and append to temporary list
        tmp_l.append(hilly_sub_table_p.get(plain_j[i].upper()))
        
    #Append the substituted P_i in the global variable
    p_blocks.append(tmp_l)
    
    print("Plaintext vector P_",j,":\n",tmp_l,sep='')

'''
This function computes the extension vector X_j for every ciphertext block C_j and
stores the result in the global list hilly_extensions.
Parameter: c_j   cipher block j
           j     only used for output
'''
def hilly_enc_extension_block_i(c_j,j):
    global hilly_extensions
    
    #Initialize list
    x = list()
    
    #Iterate cipher block j
    for d in range(len(c_j)):
        
        #Compute extension vector X
        tmp = c_j[d] / float(26)
        
        #Round up
        tmp = np.ceil(tmp)
            
        #Append to list
        x.append(int(tmp))

    #Append to global list
    hilly_extensions.append(x)
    print("Compute extension vector X_",j," = upper_bound(C1_",j,"/26): \n",x,sep='')

'''
This function computes the encrypted extension X_ENC_i for the Hilly encryption mode and stores the 
result in the global list hilly_x_enc_blocks.
Parameter: x_j    extension block j
           j      only used for output
'''
def hilly_enc_extention_enc_i(x_j,j):
    global hilly_e_matrix,hilly_x_enc_blocks
    
    #Compute encrypted extension X_ENC =  E * X_j and convert to int
    res = (np.dot(hilly_e_matrix,x_j)).astype(np.int64)
    
    #Append to global list  
    hilly_x_enc_blocks.append(res)
    
    print("Compute encrypted extension: X_ENC_",j," = E * ","X_",j,":\n",res,sep='') 
    
'''
This function uses the additional substitution table for Hilly encryption mode to substituting the encrypted extension X_ENC_i.
The result is stored in the global list hilly_cipher_blocks_extended.
Parameter: x_j    encrypted extension block
           j      indicate which block we use
'''
def hilly_enc_extention_sub_i(x_j,j):
    global hilly_sub_table_add,hilly_sub_table_add_minus,cipher_blocks,hilly_cipher_blocks_extended
    
    #Initialize variable
    c_bock_i_final = ''
    
    #Iterate encrypted extension block
    for m in range(len(x_j)):
        
        #Get first char of the cipher_blocks at position m
        tmp_s = str(cipher_blocks[j][m])
        
        #Add '-' from additional substitution table
        if x_j[m] < 0:
            
                tmp_s = tmp_s + hilly_sub_table_add_minus
        
        #Get string value at position m of encrypted extension block
        value = str(abs(x_j[m]))
           
        #Iterate string
        for i in range(len(value)):
            
            #Use additional substituion table to substituting the X_ENC block j at position i
            c = hilly_sub_table_add[int(value[i])]
            
            #Concat substituted values
            tmp_s = tmp_s + c
    
        #Concat to final string
        c_bock_i_final = c_bock_i_final + tmp_s
        
    #Append to global list
    hilly_cipher_blocks_extended.append(c_bock_i_final)
    
    
    print("After substituting C_",j,"||X_enc_",j,":\n",c_bock_i_final,sep='')

'''
This function splits the ciphertext into blocks of size n (key matrix dimension).
The result is stored in the global variable hilly_cipher_blocks_extended.
'''
def hilly_dec_split():
    global in_s,matrix_size,hilly_cipher_blocks_extended
    
    #Initialize variables
    ind = list()
    tmp_l = list()
    block_counter = 0
    
    #Iterate ciphertext and get indices of alphabetic chars
    for i in range(len(in_s)):
        if in_s[i].isalpha() == True:
            ind.append(i)
    ind.append(i)
    
    #Iterate indices of alphabetic chars and build a substring (window) for a ciphertext
    for j in range(len(ind)-1):
        
        #Window start position
        window_s = ind[j]
        
        #Window end position
        window_e = ind[j+1]
        
        #Check bounds of last window
        if window_e == len(in_s)-1:
            window_e = window_e + 1

        #We reached the block size
        if block_counter == matrix_size:
            
            #Check if we have a multiple of block size
            if len(tmp_l) == matrix_size:
                #Append to global list
                hilly_cipher_blocks_extended.append(tmp_l)
            else:
                sys.exit("Ciphertext is not a multiple of the key matrix dimension.")
            #Initialize temporary list again
            zero_l = []
            tmp_l = list(zero_l)
            
            #Reset block counter
            block_counter = 1    
        else:
            #Increment block counter
            block_counter = block_counter+1

        #Append window of ciphertext to temporary list
        tmp_l.append(in_s[window_s:window_e])
    #Check if we have a multiple of the block size
    if len(tmp_l) == matrix_size:
        #Append to global list
        hilly_cipher_blocks_extended.append(tmp_l)
    else:
        sys.exit("Ciphertext is not a multiple of the key matrix dimension.")
        
'''
This function computes the inverse key matrix K^-1 and the extension matrix E^-1.
The results are stored in the global variable key_matrix_inverse and hilly_e_matrix_inverse.
'''
def hilly_dec_gen_invert():
    global key_matrix,key_matrix_inverse,hilly_e_matrix,hilly_e_matrix_inverse
    
    #Compute inverted key matrix K^-1
    key_matrix_inverse = np.linalg.inv(key_matrix)
    
    print("Inverted key matrix K^-1:\n",key_matrix_inverse,sep='')
    

    #Compute inverted extension matrix E^-1
    hilly_e_matrix_inverse = np.linalg.inv(hilly_e_matrix)
        
    print("\nInverted extension matrix E^-1:\n",hilly_e_matrix_inverse,sep='')

'''
This function substitutes the concatenated block C_j||X_ENC_j from chars to integers for the Hilly cipher decryption mode.
The results are stored in the global lists hilly_x_enc_blocks and cipher_blocks.
Parameter: x_j   C_j concat X_ENC_j
           j     only used for output
'''   
def hilly_dec_sub_i(x_j,j):
    global sub_table_c,hilly_sub_table_add,hilly_sub_table_add_minus,hilly_x_enc_blocks,cipher_blocks
    
    print("Before substituting C_",j,"||X_ENC_",j,":\n",x_j,sep='')
    
    #Initialize lists
    c_tmp_i = list()
    x_enc_block_tmp = list()
    
    #Iterate encrypted extension block X_ENC_j
    for a in range(len(x_j)):
        #Get first char
        alpha_c = x_j[a][0]
        
        #Use substitution table for ciphertext (T_C) and get int value
        sub_int = list(sub_table_c).index(alpha_c.upper())
        
        #Append to temporary list
        c_tmp_i.append(sub_int)
        
        #Initialize variables
        x_enc_row_i = list()
        sub_final_str = ''
        
        #Iterate row of the X_ENC_j block and ignore first position
        for b in range(1,len(x_j[a])):
            #Get char at position b of the current row
            sub_c_e = x_j[a][b]
            
            #Insert minus sign from additional substitution table before substituted X_ENC
            if sub_c_e == hilly_sub_table_add_minus:
                sub_final_str = sub_final_str + '-'
                
            else:
                #Use additional substitution table and get integer value and convert to string
                sub_int_tmp = str(hilly_sub_table_add.index(sub_c_e))
                #Concat substituted values
                sub_final_str = sub_final_str + sub_int_tmp

        #Append the substituted integer value to temporary list
        x_enc_row_i.append(int(sub_final_str))
        #Append row to temporary block
        x_enc_block_tmp.append(x_enc_row_i)

    #Append to global lists
    hilly_x_enc_blocks.append(x_enc_block_tmp)
    cipher_blocks.append(c_tmp_i)
    
    print("\nAfter substitution.")
    print("\nCipher block C_",j,":\n",cipher_blocks[j],sep='')
    print("Encrypted extension block X_ENC_",j,":\n",hilly_x_enc_blocks[j],sep='')

'''
This function decrypts the encrypted extension block (X_ENC_j) for Hilly decryption mode and stores the result
in the global list hilly_extensions.
Parameter: x_j   encrypted extension block j
           j     only used for output
'''
def hilly_dec_hilly_extensions_i(x_j,j):
    global hilly_e_matrix_inverse,matrix_size,hilly_extensions
    
    #Compute: X_i = E^-1 * X_ENC_i
    ex = np.dot(hilly_e_matrix_inverse,x_j)
    
    #Round and convert as int
    ex = np.round(ex, decimals=0).astype(np.int64)
    
    print("Decrypt extension vector X_",j," = E^-1 * X_ENC_",j,":\n",np.reshape(ex, (1, matrix_size))[0],sep='')
    
    #Append to global list
    hilly_extensions.append(ex)
    
'''
This function computes C1_j for Hilly decryption mode and stores the result in the global list hilly_c1_blocks.
Parameter: x_j   extension block X_j
           c_j   cipher block C_j
             j   only used for output
'''
def hilly_dec_c1_i(x_j,c_j,j):
    global hilly_c1_blocks,matrix_size
    
    print("Compute C1_",j," = P1 - P2:\n",sep='')

    #Compute p1 = X_j * 26 
    p1 = np.dot(x_j,26)
    
    print("P1 = X_",j," * 26:\n",np.reshape(p1, (1, matrix_size))[0],sep='')
    
    #Convert C_j from list to numpy ndarray as column vector
    tmp = np.array(c_j).reshape(matrix_size,1)
    
    #Compute p2 = (26 - C_j)
    p2 = np.subtract(26,tmp)
    p2 = np.mod(p2,26)
    
    print("\nP2 = 26 - C_",j,":\n",np.reshape(p2, (1, matrix_size))[0],sep='')
    
    #Compute C1_j = (p1 - p2) 
    c1 = np.subtract(p1,p2)
    print("\nC1_",j," = P1 - P2:\n",np.reshape(c1, (1, matrix_size))[0],sep='')
    
    #Append to global list 
    hilly_c1_blocks.append(c1)
    
'''
This function decrypts the P vector block i for Hilly and uses the
plaintext substitution table (T_P) to substituting the plaintext vector to plaintext. 
The results are stored in the global lists p_blocks and plaintext_blocks.
Parameter: c1_j   c1 block j
              j   only used for output
'''
def hilly_dec_p_i(c1_j,j):
    global key_matrix_inverse,p_blocks,hilly_sub_table_p,plaintext_blocks

    #Compute P_j = K^(-1) * C_j  
    p_tmp = np.dot(key_matrix_inverse,c1_j)

    #Round and convert as int
    p_tmp = np.round(p_tmp,decimals=0).astype(np.int64)
    
    print("Compute P vector P_",j,"= K^(-1) * C1_",j,":\n",np.reshape(p_tmp, (1, matrix_size))[0],sep='')
    
    #Append decrypted P_j block to global list
    p_blocks.append(p_tmp)

    #Initialize variable
    tmp_plain = ''
    
    #Iterate P block i
    for i in range(len(p_tmp)):
        
        #Get integer value a position i
        val = p_tmp[i]
              
        #Iterate substitution table for plaintext
        for j in range(len(hilly_sub_table_p)):
            #Get alphabet value
            c = string.ascii_uppercase[j]
            #Check if integer value of p block at position i is equal with integer value from substitution table
            if val == hilly_sub_table_p.get(c):
                #Concatenate the string
                tmp_plain = tmp_plain + c

    print("\nSubstituted P vector:\n",tmp_plain,sep='')
    
    #Append to global list
    plaintext_blocks.append(tmp_plain)

'''
This function decrypts the ciphertext for the Hill and the Hilly cipher and saves the plaintext to file if wanted.
'''
def decrypt():
    global verbose_output,key_file,in_s,d_opt,matrix_size,pad_size,cipher_blocks,key_matrix_inverse,plaintext_blocks
    
    #Check if key file was used
    if key_file == False:
        sys.exit("Need key file for decryption.")

    #Check if we have an input string
    if in_s == None:
        sys.exit("No ciphertext file chosen.")
    
    if verbose_output == False:
        activate_logger()
    
    #Generate key file
    gen_key_matrix_tries()
    
    
    if verbose_output == False:
        deactivate_logger()
        print("Ciphertext found:",in_s)
        line()
        activate_logger()
    else:
        line()
        print("Ciphertext found:",in_s)
    
    #Initialize variable
    plaintext_final = ''
    
    #Find padding marker
    val = in_s.find(";")
    if val == -1:
        print("=> no padding found.")
    else:
        #Get amount of padded values
        pad_size = int(in_s[val+1:])
        
        print("=> found ",pad_size," (insecure) padded values.",sep='')
        #Save input without padding
        in_s = in_s[:val]
    line()
    #Check which cipher is used
    if d_opt != 1: #Hilly is activated
        global hilly_s,hilly_hilly_extensions,hilly_hilly_c1_blocks,hilly_hilly_cipher_blocks_extended

        #Check if s is set to 1
        if flag_hilly_s == 1:
            
            #Output 
            if verbose_output == False:
                deactivate_logger()
                print("Flag s is set. So the multiplier for the elements in (T_P) is fix and has the value 1 like in the normal Hill cipher.")
                line()
                activate_logger()
            else:    
                print("Flag s is set. So the multiplier for the elements in (T_P) is fix and has the value 1 like in the normal Hill cipher.")
                line()
            hilly_s = 1

        #Generate substitution table for plaintext and ciphertext
        hilly_gen_sub_tables()
        
        #Generate additional substitution table
        hilly_gen_sub_table_extended()
        
        #Generate extension matrix
        line()
        hilly_e_init()
        line()
        
        #Generate inverted key matrix and inverted extension matrix
        hilly_dec_gen_invert()
        line()
        
        #Split plaintext into a multiple of matrix_size (dimension of key matrix)
        hilly_dec_split()
 
        print("\nIterate ciphertext blocks..")
        line()
        for j in range(len(hilly_cipher_blocks_extended)):
            
            print("Block:",j)
            print("*****************************************************")
            
            #Get X_ENC_j block as int
            hilly_dec_sub_i(hilly_cipher_blocks_extended[j],j)
            line()
            
            #Decrypt the X_ENC_j blocks
            hilly_dec_hilly_extensions_i(hilly_x_enc_blocks[j],j)
            line()
            
            #Compute C1_i blocks
            hilly_dec_c1_i(hilly_extensions[j],cipher_blocks[j],j)
            line()
            
            #Compute P vector and get plaintext
            hilly_dec_p_i(hilly_c1_blocks[j],j)
            
            print("*****************************************************")
            
            #Concatenate plaintext blocks
            plaintext_final = plaintext_final + ''.join(k for k in plaintext_blocks[j])

    else: #Hill is activated
        global hill_sub_table_p
        
        #Generate substitution table for plain and ciphertext
        hill_gen_sub_tables()
        line()
        
        #Generate inverted key matrix
        hill_dec_gen_invert()
        line()
        
        #Split plaintext into a multiple of matrix_size (dimension of key matrix)
        c_blocks = split_it(in_s, matrix_size)
        
        print("Iterate ciphertext blocks..\n")
        for j in range(len(c_blocks)):
            
            print("Block:",j)
            print("*****************************************************")
            
            #Get C_j block as integer
            hill_dec_sub_i(c_blocks[j],j)
            
            #Compute P vector and get plaintext
            hill_dec_p_i(cipher_blocks[j],j)
            print("*****************************************************")
            
            #Concatenate plaintext blocks
            plaintext_final = plaintext_final + ''.join(k for k in plaintext_blocks[j])  
            
    #Check if padding was used
    if pad_size == 0 or pad_size == matrix_size:
        print("No remove of padding needed.")
    else:
        plaintext_final = remove_padding(plaintext_final,pad_size)
    line()
    if verbose_output == False:  
        close_logger()
    print("Plaintext:\n",plaintext_final,sep='')
            
    #Write plaintext string to file
    if path_o != None:
        line()
        save_file(plaintext_final)

'''
This function encrypts the plaintext for Hill and Hilly and saves the ciphertext to file if wanted.
'''        
def encrypt():
    global verbose_output,matrix_size,s,in_s,d_opt,p_blocks,pad_size,path_o
    
    #Start logging.
    if verbose_output == False:
        activate_logger()
    
    #Generate key file
    gen_key_matrix_tries()
    
    #Check if we have an input string
    if in_s == None:
        sys.exit("No plaintext file chosen.")
    
    line()
    
    #Output at console
    if verbose_output == False:
        deactivate_logger()
        print("Plaintext found:",in_s)
        line()
        print("Starting encryption..")
        line()
        activate_logger()
    else:
        print("Plaintext found:",in_s)
        
    line()
    print("Starting encryption..")
    line()
    
    #Check if plaintext needs padding 
    print("Check if padding is needed..")
    in_s = check_plain_padding(in_s,matrix_size)
    line()
    
    #Split plaintext into a multiple of matrix_size (dimension of key matrix)
    c = split_it(in_s,matrix_size)
    
    #Initialize variable
    cipher_str = ''
    
    #Check which cipher is used
    if d_opt != 1: #Hilly is activated
        global flag_hilly_s,hilly_s,hilly_c1_blocks,hilly_extensions,hilly_x_enc_blocks,hilly_cipher_blocks_extended
         
        #Check if s is set to 1
        if flag_hilly_s == 1:
            
            if verbose_output == False:
                deactivate_logger()
                print("Flag s is set. So the multiplier for the elements in (T_P) is fix and has the value 1 like in the normal Hill cipher.")
                line()
                activate_logger()
                
            else:
                print("Flag s is set. So the multiplier for the elements in (T_P) is fix and has the value 1 like in the normal Hill cipher.")
                line()
            hilly_s = 1
            
        #Generate substitution table for plaintext and ciphertext
        hilly_gen_sub_tables()      
        
        #Generate additional substitution table
        hilly_gen_sub_table_extended()
        line()
        
        #Generate extension matrix
        hilly_e_init()
        line()
        print("\nIterate plaintext blocks..")
        line()
        for j in range(len(c)):
            
            print("Block:",j)
            print("*****************************************************")
            
            #Get plaintext as plaintext vector P_j
            hilly_enc_sub_p_i(c[j],j)
            line()
            
            #Compute ciphertext blocks C_j and C1_j
            create_c_i(p_blocks[j],j)
            line()
            
            #Compute extension X_j
            hilly_enc_extension_block_i(hilly_c1_blocks[j],j)
            line()
            
            #Compute encrypted extension X_ENC_j
            hilly_enc_extention_enc_i(hilly_extensions[j],j)
            line()
            
            #Get encrypted extension X_ENC_j as substituted block and get ciphertext
            hilly_enc_extention_sub_i(hilly_x_enc_blocks[j],j)
            
            print("*****************************************************")
            #Concatenate plaintext blocks
            cipher_str = cipher_str + ''.join(h for h in hilly_cipher_blocks_extended[j])
    else: #Hill is activated
        global cipher_blocks
                
        #Generate substitution table for plaintext and ciphertext
        hill_gen_sub_tables()
        line()
        
        print("\nIterate plaintext blocks..")
        for j in range(len(c)):
            
            print("Block:",j)
            print("*****************************************************")
            
            #Compute plaintext vector P_j from plaintext
            hill_enc_sub_p_i(c[j],j)
            line()
            
            #Compute ciphertext C for every block j
            create_c_i(p_blocks[j],j)
            print("*****************************************************")
            
            #Concatenate ciphertext blocks
            cipher_str  = cipher_str + ''.join(x for x in cipher_blocks[j])
            line()
            
    #Check if we need padding
    if pad_size == 0 or pad_size == matrix_size:
        print("No padding of ciphertext needed.")
    else:
        print("Append (insecure) padding to ciphertext..")
        cipher_str = cipher_str + ";" + str(pad_size)
    line()
    
    #Deactivate and close logger for final output
    if verbose_output == False:  
        close_logger()
    
    print("Ciphertext:\n",cipher_str,sep='')

    #Write plaintext string to file
    if path_o != None:
        line()
        save_file(cipher_str)

def main():
    global verbose_output,matrix_size,d_opt,isKeyFile,min_val_key,max_val_key,path_o,n,key_file,in_s,flag_hilly_s,deeps,hilly_vuln,hilly_user_sub
    
    #Create arguments.
    parser = argparse.ArgumentParser(description="Hilly is an extension of the Hill cipher. This is "+version+".")

    parser.add_argument("-i", "--input", help="input file.")
    parser.add_argument("-k", "--key", help="key file.")
    parser.add_argument("-o", "--output", help="output file.")
    parser.add_argument("-a", "--addsub", help="take additional sub table for Hilly from a given file.")
    parser.add_argument("-n", "--matrix", help="set matrix dimension. Default is "+str(default_matrix_size)+".", type=int)
    parser.add_argument("-f", "--tries", help="amount of tries to generate key file.", type=int)
    parser.add_argument("-x", "--min", help="minimum int value for generating key matrix.", type=int)
    parser.add_argument("-y", "--max", help="maximum int value for generating key matrix.", type=int)
    parser.add_argument("-e", "--hill", action="store_true", help="use usual Hill cipher.")
    parser.add_argument("-s", "--default", action="store_true", help="set plaintext substitution table values for Hilly to 1 (like in normal Hill).")
    parser.add_argument("-u", "--vuln", action="store_true", help="allow insecure key (invertible det(k) mod 26 is fulfilled) for Hilly (for testing purposes only).")
    parser.add_argument("-V", "--version", action='version', version='%(prog)s '+version)
    parser.add_argument("-v", "--verbose", help="turn on multiple levels of verbosity.", action="count", default=0)
    
    
    parser.add_argument("mode",help="either encrypt (enc), decrypt (dec) or generate (gen).")

    if len(sys.argv)==1: # Abfrage VOR args=parser.parse_args(), da das parse immmer ein Argument wie mode oder -V braucht
        print(sys.argv[0], version)
        sys.exit(0)

    args = parser.parse_args()

    #Check options first [keep in mind: the following code is only reached if NOT (no arg, not -V, or not --help)]
    if args.input:#Input.
        in_s = read_f(args.input)
    if args.output:#Output.
        path_o = args.output
    if args.key:#Key file.
        key_file = read_f(args.key)
    if args.hill:#Flag which cipher is used.
        d_opt = 1
    if args.verbose:#Flag if reduced output is used.
        verbose_output = True

    if d_opt == 1:
        print("Normal Hill Cipher used.")
        line()
    else:
        print("Extended Hill Cipher (Hilly) used.")
        line()
        #Check only options for Hilly.
        if args.default:#Set s to 1.
            flag_hilly_s = 1
        if args.vuln:#Allow insecure key.
            hilly_vuln = True
        if args.addsub:#Additional sub table file.
            hilly_user_sub = read_f(args.addsub)

    #Check which mode is used.
    if args.mode == "encrypt" or args.mode == "enc":
        encrypt()
        sys.exit(0)
    elif args.mode == "decrypt" or args.mode == "dec":
        decrypt()
        sys.exit(0)
    elif args.mode == "generate" or args.mode == "gen":
        #Check options only for generate.
        if args.tries:#Amount of tries
            deeps = args.tries
        if args.matrix:#Key matrix dimension
            matrix_size = args.matrix
        if args.min:#Minimum integer value.
            min_val_key = args.min
        if args.max:#Maximum integer value.
            max_val_key = args.max
        gen_key_matrix_tries()
        
        if path_o != None:
            save_key_matrix()
        sys.exit(0)    
    else:
        print("Warning: an error occurred while parsing arguments, use --h for help.")
        sys.exit(65)
    
'''
Main function for flow control and parsing arguments.
'''
if __name__ == '__main__':
    main()
